:root {
  --md-admonition-icon--user: url('data:image/svg+xml;charset=utf-8,%3Csvg%20class%3D%22lucide%20lucide-user-round%22%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%20width%3D%2224%22%20height%3D%2224%22%20viewBox%3D%220%200%2024%2024%22%20fill%3D%22none%22%20stroke%3D%22currentColor%22%20stroke-width%3D%222%22%20stroke-linecap%3D%22round%22%20stroke-linejoin%3D%22round%22%20%3E%20%3Ccircle%20cx%3D%2212%22%20cy%3D%228%22%20r%3D%225%22%20%2F%3E%20%3Cpath%20d%3D%22M20%2021a8%208%200%200%200-16%200%22%20%2F%3E%20%3C%2Fsvg%3E');
  --md-admonition-icon--yee88: var(--md-admonition-icon--info);
  --md-admonition-color--user: #64dd17;
  --md-admonition-color--yee88: #f4a5b8;
  --md-admonition-bg-color--user: #64dd171a;
  --md-admonition-bg-color--yee88: #f4a5b81a;
}

.md-typeset .admonition.user,
.md-typeset details.user {
  background-color: var(--md-admonition-bg-color--user);
}

.md-typeset .admonition.yee88,
.md-typeset details.yee88 {
  background-color: var(--md-admonition-bg-color--yee88);
}

.md-typeset .user > .admonition-title::before,
.md-typeset .user > summary::before {
  background-color: var(--md-admonition-color--user);
  -webkit-mask-image: var(--md-admonition-icon--user);
  mask-image: var(--md-admonition-icon--user);
}

.md-typeset .user > .admonition-title::after,
.md-typeset .user > summary::after {
  color: var(--md-admonition-color--user);
}

.md-typeset .yee88 > .admonition-title::before,
.md-typeset .yee88 > summary::before {
  background-color: var(--md-admonition-color--yee88);
  -webkit-mask-image: var(--md-admonition-icon--yee88);
  mask-image: var(--md-admonition-icon--yee88);
}

.md-typeset .yee88 > .admonition-title::after,
.md-typeset .yee88 > summary::after {
  color: var(--md-admonition-color--yee88);
}
